# -*- coding: utf-8 -*-
"""
Created on Tue Jul 23 19:33:20 2019

@author: lior
"""
import numpy as np
import xarray as xr
import os
import ReadIM
from pivpy.io import parse_header, loadvec


def ReadDavis(path, time=0):
    """
    input path for files format from davis tested for im7&vc7
    out put [X Y U V mask]
    valid only for 2d piv cases
    RETURN:
     in case of images (image type=0):
              X = scaled x-coordinates
              Y = scaled y-coordinates
              U = scaled image intensities
              v=0
              MASK=0
            in case of 2D vector fields (A.IType = 1,2 or 3):
              X = scaled x-coordinates
               Y = scaled y-coordinates
               U = scaled vx-components of vectors
               V = scaled vy-components of vectors

    """
    # you need to add clear to prevent data leaks
    buff, vatts = ReadIM.extra.get_Buffer_andAttributeList(path)
    v_array, buff1 = ReadIM.extra.buffer_as_array(buff)
    nx = buff.nx
    # nz = buff.nz #flake8 claims it's not used
    ny = buff.ny
    # set data range:
    baseRangeX = np.arange(nx)
    baseRangeY = np.arange(ny)
    # baseRangeZ = np.arange(nz) #flake8 recognized it's not used
    lhs1 = (
        baseRangeX + 0.5
    ) * buff.vectorGrid * buff.scaleX.factor + buff.scaleX.offset  # x-range
    lhs2 = (
        baseRangeY + 0.5
    ) * buff.vectorGrid * buff.scaleY.factor + buff.scaleY.offset  # y-range
    lhs3 = 0
    lhs4 = 0
    mask = 0
    if buff.image_sub_type <= 0:  # grayvalue image format
        lhs3 = v_array[0, :, :]
        lhs4 = v_array[1, :, :]
        Im = xr.DataArray(
            v_array,
            dims=("frame", "z", "x"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "frame": [0, 1]},
        )
        data = xr.Dataset({"Im": Im})

    elif buff.image_sub_type == 2:  # simple 2D vector format: (vx,vy)
        # Calculate vector position and components
        [lhs1, lhs2] = np.meshgrid(lhs1, lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = v_array[0, :, :] * buff.scaleI.factor + buff.scaleI.offset
        lhs4 = v_array[1, :, :] * buff.scaleI.factor + buff.scaleI.offset
        if buff.scaleY.factor < 0.0:
            lhs4 = -lhs4
        lhs3 = lhs3[:, :, np.newaxis]
        lhs4 = lhs4[:, :, np.newaxis]
        u = xr.DataArray(
            lhs3,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        v = xr.DataArray(
            lhs4,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        data = xr.Dataset({"u": u, "v": v})
    #    	plt.quiver(lhs1,lhs2,lhs3,lhs4);
    elif buff.image_sub_type == 3 or buff.image_sub_type == 1:
        # normal 2D vector format + peak: sel+4*(vx,vy) (+peak)
        # Calculate vector position and components
        [lhs1, lhs2] = np.meshgrid(lhs1, lhs2)
        #    lhs1=np.transpose(lhs1)
        #    lhs2=np.transpose(lhs2)
        lhs3 = lhs1 * 0
        lhs4 = lhs2 * 0
        # Get choice
        maskData = v_array[0, :, :]
        # Build best vectors from choice field
        for i in range(5):
            mask = maskData == (i + 1)
            if i < 4:  # get best vectors
                dat = v_array[2 * i + 1, :, :]
                lhs3[mask] = dat[mask]
                dat = v_array[2 * i + 2, :, :]
                lhs4[mask] = dat[mask]
            else:  # get interpolated vectors
                dat = v_array[7, :, :]
                lhs3[mask] = dat[mask]
                dat = v_array[8, :, :]
                lhs4[mask] = dat[mask]
        lhs3 = lhs3 * buff.scaleI.factor + buff.scaleI.offset
        lhs4 = lhs4 * buff.scaleI.factor + buff.scaleI.offset
        # Display vector field
        if buff.scaleY.factor < 0.0:
            lhs4 = -1 * lhs4
        mask = maskData == 0
        lhs3 = lhs3[:, :, np.newaxis]
        lhs4 = lhs4[:, :, np.newaxis]
        maskData = maskData[:, :, np.newaxis]
        u = xr.DataArray(
            lhs3,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        v = xr.DataArray(
            lhs4,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        chc = xr.DataArray(
            maskData,
            dims=("z", "x", "t"),
            coords={"x": lhs1[0, :], "z": lhs2[:, 0], "t": [time]},
        )
        data = xr.Dataset({"u": u, "v": v, "chc": chc})
    data.attrs["Info"] = ReadIM.extra.att2dict(vatts)
    # clean memory
    ReadIM.DestroyBuffer(buff1)
    del buff1
    ReadIM.DestroyBuffer(buff)
    del buff
    ReadIM.DestroyAttributeListSafe(vatts)
    del vatts
    return data


def load_directory(path, basename=""):
    """
    load_directory (path)

    Loads all the .VEC files in the directory into a single
    xarray dataset with variables and units added as attributes

    Input:
        directory : path to the directory with .vec files

    Output:
        data : xarray DataSet with dimensions: x,y,t and
               data arrays of u,v,
               attributes of variables and units


    See more: loadvec
    """
    files = [f for f in os.listdir(path) if f.endswith(".vc7")]
    variables, units, rows, cols, dt, frame = parse_header(files[0])

    data = []
    for i, f in enumerate(files):
        data.append(
            loadvec(f, rows, cols, variables, units, dt, frame + i - 1)
        )

    combined = xr.concat(data, dim="t")
    combined.attrs["variables"] = variables
    combined.attrs["units"] = units
    combined.attrs["dt"] = dt
    combined.attrs["files"] = files
    return combined


path = "C:\\Users\\lior\\Documents\\ibrrTau\\timeDependedVecMaps"
files = [f for f in os.listdir(path) if f.endswith(".vc7")]
data = []
data.append(ReadDavis(path + "\\" + files[-1], 1))
data.append(ReadDavis(path + "\\" + files[-2], 2))
combined = xr.concat(data, dim="t")
# x =x.flatten()
# y =y.ravel()
# u =u.ravel()
# v =v.ravel()
# plt.quiver(x,y,u,v)
# plt.imshow(u)
#    u = xr.DataArray(u,dims=('y','x'),coords={'x':x[0,:],'y':y[:,0]})



# -*- coding: utf-8 -*-
"""
Various plots, mostly wraping xarray.plot 

"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, FFMpegWriter
import xarray as xr
import os
from pivpy.io import POS_UNITS, VEL_UNITS
from typing import List
import warnings


def quiver(
    data: xr.DataArray,
    arrScale: float=25.0,
    threshold: float=None,
    nthArr: int=1,
    aspectratio: str="equal",
    colorbar: bool=False,
    colorbar_orient: str="vertical",
    units: List=[],
    streamlines: bool=False
):
    """
    Generates a quiver plot of a 'data' xarray DataArray object (single frame
    from a dataset)
    Inputs:
        data - xarray DataArray of the type defined in pivpy, one of the
        frames in the Dataset
            selected by default using .isel(t=0)
        threshold - values above the threshold will be set equal to threshold
        arrScale - use to change arrow scales
        nthArr - use to plot only every nth arrow from the array
        contourLevels - use to specify the maximum value (abs) of contour plots
        colorbar - True/False wether to generate a colorbar or not
        logscale - if true then colorbar is on log scale
        aspectratio - set auto or equal for the plot's apearence
        colorbar_orient - 'horizontal' or 'vertical' orientation of the colorbar
        (if colbar is True)
    Outputs:
        none
    Usage:
        graphics.quiver(data, arrScale = 0.2, threshold = Inf, n)
    """

    data = dataset_to_array(data)


    pos_units = data.x.attrs["units"] if len(units)==0 else units[0]
    vel_units = data.u.attrs["units"] if len(units)==0 else units[2]

    # clip data to the threshold
    if threshold is not None:
        data["u"] = xr.where(data["u"] > threshold, threshold, data["u"])
        data["v"] = xr.where(data["v"] > threshold, threshold, data["v"])

    data['s'] = np.sqrt(data["u"] ** 2 + data["v"] ** 2)

    if len(plt.get_fignums()) == 0:  # if no figure is open
        fig, ax = plt.subplots()  # open a new figure
    else:
        fig = plt.gcf()
        ax = fig.gca()

    # quiver itself
    Q = data.plot.quiver(
            x='x',
            y='y',
            u='u',
            v='v',
            hue='s',
            units='width',
            scale=np.max(data['s'].values * arrScale),
            headwidth=2,
            ax=ax,
            )

    if colorbar is False:
        # cbar = fig.colorbar(Q, shrink=0.9, orientation=colbar_orient)
        cb = Q.colorbar
        cb.remove()
        plt.draw()
    else:
        if colorbar_orient == 'horizontal':
            cb = Q.colorbar
            cb.remove()
            cb = fig.colorbar(Q, orientation=colorbar_orient, ax=ax)


    if streamlines:  # contours or streamlines
        strm = data.plot.streamplot(
            x='x',
            y='y',
            u='u',
            v='v',
            hue='s',
            cmap="hot", 
            linewidth=1,
            ax=ax,
        )
        strm.colorbar.remove()

        # if colorbar:
        #     cbar = fig.colorbar(
        #         strm, 
        #         orientation=colorbar_orient, 
        #         fraction=0.1,
        #     )
        #     cbar.set_label(r"$ V \, (" + vel_units + r")$")

    ax.set_xlabel(f"x ({pos_units})")
    ax.set_ylabel(f"y ({pos_units})")
    ax.set_aspect(aspectratio)
    # ax.invert_yaxis()

    return fig, ax


def histogram(data, normed=False):
    """
    this function will plot a normalized histogram of
    the velocity data.
    Input:
        data : xarray DataSet with ['u','v'] attrs['units']
        normed : (optional) default is False to present normalized
        histogram
    """

    u = np.asarray(data.u).flatten()
    v = np.asarray(data.v).flatten()

    f, ax = plt.subplots(2)

    ax[0].hist(u, bins = np.int32(np.sqrt(len(u)) * 0.5), density=normed)
    ax[0].set_xlabel(f"u ({data.u.attrs['units']})")

    ax[1] = plt.subplot2grid((2, 1), (1, 0))
    ax[1].hist(v, bins = np.int32(np.sqrt(len(v) * 0.5)), density=normed)
    ax[1].set_xlabel(f"v ({data.v.attrs['units']})")
    plt.tight_layout()
    return f, ax


def contour_plot(
    data: xr.DataArray,
    threshold: float=None,
    contourLevels: List[float]=[],
    colorbar: bool=False,
    logscale: bool=False,
    aspectratio: str="equal",
    units: List[str] = [],
):
    """ contourf ajusted for the xarray PIV dataset, creates a
        contour map for the data['w'] property.
        Input:
            data : xarray PIV DataArray, converted automatically using
            .isel(t=0)
            threshold : a threshold value, default is None (no data clipping)
            contourLevels : number of contour levels, default is None
            colbar : None (hide), 'horizontal', or 'vertical'
            logscale : boolean (True is default) create in linear/log scale
            aspectration : string, 'equal' is the default
    """

    data = dataset_to_array(data)    

    if "w" not in data.var():
        data = data.piv.vec2scal("ke")
        
    if threshold is not None:
        data["w"] = xr.where(data["w"] > threshold, threshold, data["w"])


    f, ax = plt.subplots()
    # data.plot.contourf(x='x',y='y',row='y',col='x', ax=ax)

    if len(contourLevels) == 0:
        levels = np.linspace(
            np.min(data["w"].values), 
            np.max(data["w"].values), 
            10,
        )
    else:
        levels = contourLevels  # vector of levels to set

    if logscale:
        data["w"] = np.abs(data["w"])

        c = data["w"].plot.contourf(
            x='x',
            y='y',
            levels=levels,
            cmap=plt.get_cmap("RdYlBu"),
            norm=plt.colors.LogNorm(),
            ax=ax,
        )
    else:
        c = data["w"].plot.contourf(
            x='x',
            y='y',
            levels=levels,
            cmap=plt.get_cmap("RdYlBu"),
            ax=ax,
        )

    if not colorbar:
        # cbar = c.colorbar(c, orientation=colbar)
        c.colorbar.remove()
        # cbar.set_label(propUnits)

    ax.set_aspect(aspectratio)

    return f, ax


def showf(data, property="ke", **kwargs):
    """
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and variables u,v and maybe w (scalar)
    """
    fig, ax = showscal(data, property=property, **kwargs)
    fig, ax = quiver(data,**kwargs)


def showscal(data, property="ke", **kwargs):
    """
    showf(data, var, units)
    Arguments:
        data : xarray.DataSet that contains dimensions of t,x,y
               and a variable w (scalar)
    """
    data = data.piv.vec2scal(property=property)
    fig, ax = contour_plot(data)
    return fig, ax


def animate(data, arrowscale=1, savepath=None):
    """ animates the quiver plot for the dataset (multiple frames)
    Input:
        data : xarray PIV type of DataSet
        arrowscale : [optional] integer, default is 1
        savepath : [optional] path to save the MP4 animation, default is None

    Output:
        if savepath is None, then only an image display of the animation
        if savepath is an existing path, a file named im.mp4 is saved

    """
    X, Y = np.meshgrid(data.x, data.y)
    X = X.T
    Y = Y.T
    U, V = data.u[:, :, 0], data.v[:, :, 0]  # first frame
    fig, ax = plt.subplots(1, 1)
    M = np.sqrt(U ** 2 + V ** 2)

    Q = ax.quiver(
        X[::3, ::3],
        Y[::3, ::3],
        U[::3, ::3],
        V[::3, ::3],
        M[::3, ::3],
        units="inches",
        scale=arrowscale,
    )

    cb = plt.colorbar(Q)

    units = data.attrs["units"]

    cb.ax.set_ylabel("velocity (" + units[2] + ")")

    text = ax.text(
        0.2,
        1.05,
        "1/" + str(len(data.t)),
        ha="center",
        va="center",
        transform=ax.transAxes,
    )

    def update_quiver(num, Q, data, text):
        U, V = data.u[:, :, num], data.v[:, :, num]

        M = np.sqrt(U[::3, ::3] ** 2 + V[::3, ::3] ** 2)
        Q.set_UVC(U[::3, ::3], V[::3, ::3], M[::3, ::3])
        text.set_text(str(num + 1) + "/" + str(len(data.t)))
        return Q

    anim = FuncAnimation(
        fig,
        update_quiver,
        fargs=(Q, data, text),
        frames=len(data.t),
        blit=False,
    )
    mywriter = FFMpegWriter()
    if savepath:
        p = os.getcwd()
        os.chdir(savepath)
        anim.save("im.mp4", writer=mywriter)
        os.chdir(p)
    else:
        anim.save("im.mp4", writer=mywriter)


def dataset_to_array(data:xr.Dataset, t:int=0):
    """ converts xarray Dataset to array """
    if "t" in data.dims:
        warnings.warn("Warning: function for a single frame, using the first \
               frame, supply data.isel(t=N)")
        return data.isel(t=t)
    else:
        return data


# -*- coding: utf-8 -*-
"""
Created on Sun May 24 22:02:49 2015

@author: Ron, Alex
"""
import numpy as np

# from scipy.stats import norm
# from scipy.ndimage.filters import gaussian_filter
# from scipy.ndimage.filters import median_filter
import xarray as xr
# from pivpy import graphics
from scipy.ndimage import gaussian_filter as gf

""" learn from this example


import xarray as xr


@xr.register_dataset_accessor('geo')
class GeoAccessor(object):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
        self._center = None

    @property
    def center(self):
        " Return the geographic center point of this dataset."
        if self._center is None:
            # we can use a cache on our accessor objects, because accessors
            # themselves are cached on instances that access them.
            lon = self._obj.latitude
            lat = self._obj.longitude
            self._center = (float(lon.mean()), float(lat.mean()))
        return self._center

    def plot(self):
        " Plot data on a map."
        return 'plotting!'


    In [1]: ds = xr.Dataset({'longitude': np.linspace(0, 10),
   ...:                  'latitude': np.linspace(0, 20)})
   ...:

In [2]: ds.geo.center
Out[2]: (10.0, 5.0)

In [3]: ds.geo.plot()
Out[3]: 'plotting!'

"""


@xr.register_dataset_accessor("piv")
class PIVAccessor(object):
    def __init__(self, xarray_obj):
        """
        Arguments:
            data : xarray Dataset:
            x,y,t are coordinates
            u,v,chs are the data arrays

        We add few shortcuts (properties):
            data.piv.average is the time average (data.mean(dim='t'))
            data.piv.dt is the shortcut to get $\\Delta t$
            data.piv.vorticity
            data.piv.tke
            data.piv.shear

        and a few methods:
            data.piv.vec2scal()
            data.piv.pan
            data.piv.rotate

        """
        self._obj = xarray_obj
        self._average = None
        self._dt = None

    @property
    def average(self):
        """ Return the mean flow field ."""
        if self._average is None:  # only first time
            self._average = self._obj.mean(dim="t")
            self._average.attrs = self._obj.attrs  # we need units in quiver
            self._average.assign_coords({"t": 0})

        return self._average

    def crop(self, crop_vector=[None, None, None, None]):
        """ crop number of rows, cols from either side of the vector fields
        Input:
            self : xarray Dataset
            crop_vector : [xmin,xmax,ymin,ymax] is a list of values crop
                            the data, defaults are None
        Return:
            same object as the input
        """
        xmin, xmax, ymin, ymax = crop_vector

        xmin = self._obj.x.min() if xmin is None else xmin
        xmax = self._obj.x.max() if xmax is None else xmax
        ymin = self._obj.y.min() if ymin is None else ymin
        ymax = self._obj.y.max() if ymax is None else ymax

        self._obj = self._obj.sel(x=slice(xmin, xmax), y=slice(ymin, ymax))

        return self._obj

    def pan(self, dx=0.0, dy=0.0):
        """ moves the field by dx,dy in the same units as x,y """
        self._obj = self._obj.assign_coords(
            {"x": self._obj.x + dx, "y": self._obj.y + dy}
        )
        # self._obj['x'] += dx
        # self._obj['y'] += dy
        return self._obj

    def filterf(self):
        """Gaussian filtering of velocity """

        self._obj["u"] = xr.DataArray(
            gf(self._obj["u"].values, [1, 1, 0]), dims=("x", "y", "t")
        )
        self._obj["v"] = xr.DataArray(
            gf(self._obj["v"].values, [1, 1, 0]), dims=("x", "y", "t")
        )

        return self._obj

    def __add__(self, other):
        """ add two datasets means that we sum up the velocities, assume
        that x,y,t,dt are all identical
        """
        self._obj["u"] += other._obj["u"]
        self._obj["v"] += other._obj["v"]
        return self._obj

    def __sub__(self, other):
        """ add two datasets means that we sum up the velocities, assume
        that x,y,t,dt are all identical
        """
        self._obj["u"] -= other._obj["u"]
        self._obj["v"] -= other._obj["v"]
        return self._obj

    def vorticity(self):
        """ calculates vorticity of the data array (at one time instance) and
        adds it to the attributes

        Input:
            xarray with the variables u,v and dimensions x,y

        Output:
            xarray with the estimated vorticity as a scalar field with
            same dimensions

        """

        self._obj["w"] = (
            self._obj["v"].differentiate("x")
            - self._obj["u"].differentiate("y")
        )

        self._obj["w"].attrs["units"] = "1/dt"
        self._obj["w"].attrs["standard_name"] = "vorticity"

        return self._obj

    def strain(self):
        """ calculates strain rate, du/dx^2 + dv/dy^2 + 1/2 (du/dy+dv/dx)^2
        Input:
            xarray with the variables u,v and dimensions x,y
        Output:
            xarray with the estimated shear as a scalar field data['w']
        """
        ux = self._obj["u"].differentiate("x")
        uy = self._obj["u"].differentiate("y")
        vx = self._obj["v"].differentiate("x")
        vy = self._obj["v"].differentiate("y")

        self._obj["w"] = ux ** 2 + vy ** 2 + 0.5 * (uy + vx) ** 2
        self._obj["w"].attrs["units"] = "1/dt"
        self._obj["w"].attrs["standard_name"] = "strain"

        return self._obj

    def divergence(self):
        """ calculates shear of the data array (single frame)
        Input:
            xarray with the variables u,v and dimensions x,y
        Output:
            xarray with the estimated shear as a scalar field data['w']
        """
        ux, _ = np.gradient(
            self._obj["u"], self._obj["x"], self._obj["y"], axis=(0, 1)
        )
        _, vy = np.gradient(
            self._obj["v"], self._obj["x"], self._obj["y"], axis=(0, 1)
        )

        if "t" in self._obj.coords:
            self._obj["w"] = (("x", "y", "t"), vy + ux)
        else:
            self._obj["w"] = (("x", "y"), vy + ux)


        self._obj["w"].attrs["units"] = "1/dt"
        self._obj["w"].attrs["standard_name"] = "divergence"

        return self._obj

    def acceleration(self):
        """ calculates material derivative or acceleration of the
        data array (single frame)

        Input:
            xarray with the variables u,v and dimensions x,y

        Output:
            xarray with the estimated acceleration as a scalar field data['w']

        """
        ux = self._obj["u"].differentiate("x")
        uy = self._obj["u"].differentiate("y")
        vx = self._obj["v"].differentiate("x")
        vy = self._obj["v"].differentiate("y")

        ax = self._obj["u"] * ux + self._obj["v"] * uy
        ay = self._obj["u"] * vx + self._obj["v"] * vy

        self._obj["w"] = xr.DataArray(
            np.sqrt(ax ** 2 + ay ** 2), dims=["x", "y", "t"]
        )

        self._obj["w"].attrs["units"] = "1/dt"
        self._obj["w"].attrs["standard_name"] = "acceleration"

        return self._obj

    def kinetic_energy(self):
        """ estimates turbulent kinetic energy """
        self._obj["w"] = self._obj["u"] ** 2 + self._obj["v"] ** 2
        self._obj["w"].attrs["units"] = "(m/s)^2"
        self._obj["w"].attrs["standard_name"] = "kinetic_energy"
        return self._obj

    def tke(self):
        """ estimates turbulent kinetic energy """
        if len(self._obj.t) < 2:
            raise ValueError(
                "TKE is not defined for a single vector field, \
                              use .piv.kinetic_energy()"
            )
        
        new_obj = self._obj.copy()
        new_obj -= new_obj.mean(dim="t")
        new_obj["w"] = new_obj["u"] ** 2 + new_obj["v"] ** 2
        new_obj["w"].attrs["units"] = "(m/s)^2"
        new_obj["w"].attrs["standard_name"] = "TKE"

        return new_obj

    def fluct(self):
        """ returns fluctuations as a new dataset """

        if len(self._obj.t) < 2:
            raise ValueError(
                "fluctuations cannot be defined for a \
                              single vector field, use .piv.ke()"
            )

        new_obj = self._obj.copy()
        new_obj -= new_obj.mean(dim="t")

        new_obj["u"].attrs["standard_name"] = "fluctation"
        new_obj["v"].attrs["standard_name"] = "fluctation"

        return new_obj

    def reynolds_stress(self):
        """ returns fluctuations as a new dataset """

        if len(self._obj.t) < 2:
            raise ValueError(
                "fluctuations cannot be defined for a \
                              single vector field, use .piv.ke()"
            )

        new_obj = self._obj.copy()
        new_obj -= new_obj.mean(dim="t")

        new_obj["w"] = -1 * new_obj["u"] * new_obj["v"] # new scalar
        self._obj["w"] = new_obj["w"].mean(dim="t") # reynolds stress is -\rho < u' v'>
        self._obj["w"].attrs["standard_name"] = "Reynolds_stress"

        return self._obj

    def rms(self):
        """ Root mean square """
        self._obj = tke()
        self._obj["w"] = np.sqrt(self._obj["w"])
        self._obj["w"].attrs["standard_name"] = "rms"
        self._obj["w"].attrs["units"] = "m/s"

    def vec2scal(self, property="curl"):
        """ creates a dataset of scalar values on the same
        dimensions and coordinates as the vector dataset
        Agruments:
            data : xarray.DataSet with u,v on t,x,y grid
            property: str, one of the propertes from the list
                'vorticity','kinetic_energy', 'tke', 'curl','strain'
        Returns:
            scalar_data : xarray.Dataset w on t,x,y grid
            'w' represents one of the following properties:
                - 'curl' or 'rot' - vorticity

        """
        # replace few common names
        property = "vorticity" if property == "curl" else property
        property = "kinetic_energy" if property == "ken" else property
        property = "kinetic_energy" if property == "ke" else property
        property = "vorticity" if property == "vort" else property
 
        method = getattr(self, str(property))

        self._obj = method()

        return self._obj

    def __mul__(self, scalar):
        """
        multiplication of a velocity field by a scalar (simple scaling)
        """
        self._obj["u"] *= scalar
        self._obj["v"] *= scalar
        if 'w' in self._obj.var():
            self._obj['w'] += scalar

        return self._obj

    def __div__(self, scalar):
        """
        multiplication of a velocity field by a scalar (simple scaling)
        """
        self._obj["u"] /= scalar
        self._obj["v"] /= scalar

        return self._obj

    def set_dt(self, dt):
        self._obj.attrs["delta_t"] = dt
        return self._obj

    def set_scale(self, scale=1.0):
        for var in ['x','y','u','v']:
            self._obj[var] = self._obj[var]*scale

        return self._obj

    def set_tUnits(self, tUnits):
        self._obj.attrs["tUnits"] = tUnits

    def rotate(self, theta=0.0):
        """
        use this method in order to rotate the data
        by theta degrees in the clockwise direction
        in the present form of using xarray with coordinates
        it can only work for the cases with equal size along
        x and y
        """

        theta = theta / 360.0 * 2 * np.pi

        xi = self._obj.x * np.cos(theta) + self._obj.y * np.sin(theta)
        eta = self._obj.y * np.cos(theta) - self._obj.x * np.sin(theta)
        Uxi = self._obj.u * np.cos(theta) + self._obj.v * np.sin(theta)
        Ueta = self._obj.v * np.cos(theta) - self._obj.u * np.sin(theta)

        self._obj["x"] = xi
        self._obj["y"] = eta
        self._obj["u"] = Uxi
        self._obj["v"] = Ueta

        if "theta" in self._obj:
            self._obj["theta"] += theta
        else:
            self._obj["theta"] = theta

        return self._obj

    @property
    def dt(self):
        """ receives the dt from the set """
        if self._dt is None:
            self._dt = self._obj.attrs["delta_t"]
        return self._dt

    def quiver(self, **kwargs):
        """ graphics.quiver() as a property """
        from pivpy.graphics import quiver
        fig, ax = quiver(self._obj, **kwargs)
        return fig, ax

    def streamplot(self, **kwargs):
        """ graphics.quiver(streamlines=True) """
        from pivpy.graphics import quiver
        quiver(self._obj, streamlines=True, **kwargs)

    def showf(self, **kwargs):
        """ method for graphics.showf """
        from pivpy.graphics import showf as gshowf
        gshowf(self._obj, **kwargs)

    def showscal(self, **kwargs):
        """ method for graphics.showscal """
        from pivpy.graphics import showscal
        showscal(self._obj, **kwargs)

    # @property
    # def vel_units(self):
    #     " Return the geographic center point of this dataset."
    #     if self._vel_units is None:
    #         self._vel_units = self._obj.attrs.l_units + '/' + \
    #                           self._obj.attrs.t_units
    #     return self._vel_units



# -*- coding: utf-8 -*-

"""
Contains functions for reading flow fields in various formats
"""

from multiprocessing.dummy import Array
from socketserver import DatagramRequestHandler
import numpy as np
import xarray as xr
import pandas as pd
from glob import glob
import os
import pathlib
import re
import warnings
from typing import List, Tuple, Dict, Any
from numpy.typing import ArrayLike

import warnings

try:
    from lvreader import read_buffer
except: 
    warnings.warn('lvreader is not installed')


# Defaults
POS_UNITS: str = "pix" # or mm, m, after scaling
TIME_UNITS: str = "frame" # "frame" if not scaled, can become 'sec' or 'msec', 'usec'
# after scaling can be m/s, mm/s
VEL_UNITS: str =  POS_UNITS # default is displacement in pix
DELTA_T: np.float64 = 0.0 # default is 0. i.e. uknown, can be any float value

def sorted_unique(
    a: ArrayLike
) -> ArrayLike:
    b, c = np.unique(a, return_counts=True)
    out = b[np.argsort(-c)]
    return out

def set_default_attrs(dataset: xr.Dataset)-> xr.Dataset:
    """ Defines default attributes:

    # xr.DataSet.x.attrs["units"] = POS_UNITS
    POS_UNITS: str = "pix" # or mm, m, after scaling


    # None if not known, can become 'sec' or 'msec', 'usec'
    # the time units are for the sequence of PIV realizations
    # useful for animations of the DataSet and averaging
    # xr.DataSet.t.attrs["units"] = TIME_UNITS
    TIME_UNITS: str = None 

    # after scaling can be m/s, mm/s, default is POS_UNITS
    # xr.DataSet.u.attrs["units"] = VEL_UNITS
    VEL_UNITS: str =  POS_UNITS

    # attribute of the xr.DataSet, defines things for the 
    # single flow realization, frame A -> DELTA_T -> frame B
    # xr.DataSet.attrs["delta_t"] = DELTA_T 
    DELTA_T: float = None # default is unknown, can be
    
    """

    dataset.x.attrs["units"] = POS_UNITS
    dataset.y.attrs["units"] = POS_UNITS
    dataset.u.attrs["units"] = VEL_UNITS
    dataset.v.attrs["units"] = VEL_UNITS
    dataset.t.attrs["units"] = TIME_UNITS
    dataset.attrs["delta_t"] = DELTA_T
    dataset.attrs["files"] = []

    return dataset

def create_sample_field(
        rows: int=5,
        cols: int=8,
        grid: List=[32,16],
        frame: int=0, 
        noise_sigma: float=1.0
        ) -> xr.Dataset:
    """ 
    creates a sample Dataset for the tests 

    rows - number of points along vertical coordinate, 
           corresponds to 'y'
    cols - number of grid points along horizontal coordinate, 'x'
    grid - spacing between vectors in two directions (x,y)
    frame - frame number
    noise_sigma - strength of Gaussian noise to add

    Returns:
        xarray.Dataset()

Usage:
    io.create_sample_field(
            rows=3,
            cols=6,
            grid=[32,16],
            frame=0,
            noise_sigma=0.1
        )

"""

    x = np.arange(grid[0], (cols + 1) * grid[0], grid[0])
    y = np.arange(grid[1], (rows + 1) * grid[1], grid[1])

    xm, ym = np.meshgrid(x, y)
    u = np.ones_like(xm) + \
        np.linspace(0.0, 10.0, cols) +\
        noise_sigma * np.random.randn(1, cols)    
    v = (
        np.zeros_like(ym)
        + np.linspace(-1.0, 1.0, rows).reshape(rows, 1)
        + noise_sigma * np.random.randn(rows, 1)
    )

    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = np.ones_like(u)

    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    dataset = xr.Dataset({"u": u, "v": v, "chc": chc})
    dataset = set_default_attrs(dataset)

    return dataset


def create_sample_Dataset(
        n_frames: int=5,
        rows: int=5,
        cols: int=3, 
        noise_sigma: float=0.0
        )-> xr.Dataset:
    """ using create_sample_field that has random part in it, create
    a sample Dataset of length 'n' """

    dataset = []
    for i in range(n_frames):
        dataset.append(
            create_sample_field(
                rows=rows, 
                cols=cols, 
                frame=i, 
                noise_sigma=noise_sigma)
            )

    combined = xr.concat(dataset, dim="t")
    combined = set_default_attrs(combined)

    return combined

def create_uniform_strain():
    return create_sample_field(noise_sigma=0.0)

    
def from_arrays( x: ArrayLike, 
                 y: ArrayLike, 
                 u: ArrayLike, 
                 v: ArrayLike, 
                 mask: np.array,
                 frame: int=0):
    """
        from_arrays(x,y,u,v,mask,frame=0)
        creates an xArray Dataset from 5 two-dimensional Numpy arrays
        of x,y,u,v and mask

        Input:
            x,y,u,v,mask = Numpy floating arrays, all the same size
        Output:
            dataset is a xAarray Dataset, see xarray for help
    """
    # create dataset structure of appropriate size
    dataset = create_sample_field(rows=x.shape[0], cols=x.shape[1],frame=frame)
    # assign arrays
    dataset["x"] = x[0, :]
    dataset["y"] = y[:, 0]
    dataset["u"] = xr.DataArray(u[:, :, np.newaxis], dims=("y", "x", "t"))
    dataset["v"] = xr.DataArray(v[:, :, np.newaxis], dims=("y", "x", "t"))
    dataset["chc"] = xr.DataArray(mask[:, :, np.newaxis], dims=("y", "x", "t"))
    dataset = set_default_attrs(dataset)

    return dataset

def from_df(
    df: pd.DataFrame, 
    frame: int=0, 
    filename: str=None):
    """
        from_df(x,y,u,v,mask,frame=0)
        creates an xArray Dataset from pandas datasetframe with 5 columns

        Read the .txt files faster with pandas read_csv()

        %%time
        df = pd.read_csv(files_list[-1],delimiter='\t', 
                        names = ['x','y','u','v','mask'],header=0)
        from_df(df,filename=files_list[-1])

        is 3 times faster than the load_txt


        Input:
            x,y,u,v,mask = Numpy floating arrays, all the same size
        Output:
            dataset is a xAarray Dataset, see xarray for help
    """
    d = df.to_numpy()

    x = sorted_unique(d[:, 0])
    y = sorted_unique(d[:, 1])
    if d.shape[1] < 5: # not always there's a mask
        tmp = np.ones((d.shape[0], 5))
        tmp[:,:-1] = d
        d = tmp

    d = d.reshape(len(y), len(x), 5)  # .transpose(1, 0, 2)

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]

    # extend dimensions
    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = chc[:, :, np.newaxis]


    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    dataset = xr.Dataset({"u": u, "v": v, "chc": chc})
    dataset = set_default_attrs(dataset)
    if filename is not None:
        dataset.attrs["files"].append(str(filename))


    return dataset

def load_vec(
    filename: pathlib.Path,
    rows: int=None,
    cols: int=None,
    dt: float=None,
    frame: int=0,
)-> xr.Dataset:
    """
        load_vec(filename,rows=rows,cols=cols)
        Loads the VEC file (TECPLOT format by TSI Inc.),
        OpenPIV VEC or TXT formats
        Arguments:
            filename : file name, expected to have a header and 5 columns
            rows, cols : number of rows and columns of a vector field,
            if None, None, then parse_header is called to infer the number
            written in the header
            DELTA_T : time interval (default is None)
            frame : frame or time marker (default is None)
        Output:
            dataset is a xAarray Dataset, see xarray for help
    """
    if rows is None or cols is None:
        variables, units, rows, cols, dt, frame,_ = parse_header(filename)

    if rows is None:  # means no headers
        d = np.genfromtxt(filename, usecols=(0, 1, 2, 3, 4))
        x = sorted_unique(d[:, 0])
        y = sorted_unique(d[:, 1])
        d = d.reshape(len(y), len(x), 5)  # .transpose(1, 0, 2)
    else:
        # d = np.genfromtxt(
        #     filename, skiprows=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        # ).reshape(rows, cols, 5)
        d = np.genfromtxt(
            filename, skip_header=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        ).reshape(cols, rows, 5)
        x = d[:, :, 0][0, :]
        y = d[:, :, 1][:, 0]

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]

    # extend dimensions
    u = u[:, :, np.newaxis]
    v = v[:, :, np.newaxis]
    chc = chc[:, :, np.newaxis]

    u = xr.DataArray(
        u, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    v = xr.DataArray(
        v, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )
    chc = xr.DataArray(
        chc, dims=("y", "x", "t"), coords={"x": x, "y": y, "t": [frame]}
    )

    dataset = xr.Dataset({"u": u, "v": v, "chc": chc})

    dataset = set_default_attrs(dataset)
    if filename is not None:
        dataset.attrs["files"].append(str(filename))
    if dt is not None:
        dataset.attrs["delta_t"] = dt

    return dataset

def load_vc7(
    filename: pathlib.Path,
    frame: int=0,
)-> xr.Dataset:
    """
        load_vc7(filename) or load_vc7(filename, frame=0)
        Loads the vc7 file using Lavision lvreader package,
        Arguments:
            filename : file name, pathlib.Path
        Output:
            dataset : xarray.Dataset
    """
    buffer = read_buffer(str(filename))
    data = buffer[0] # first component is a vector frame
    plane = 0 # don't understand the planes issue, simple vc7 is 0

    u = data.components["U0"][plane]
    v = data.components["V0"][plane]

    mask = np.logical_not(data.masks[plane] & data.enabled[plane])
    u[mask] = 0.0
    v[mask] = 0.0

    # scale
    u = data.scales.i.offset + u*data.scales.i.slope
    v = data.scales.i.offset + v*data.scales.i.slope

    x = np.arange(u.shape[1])
    y = np.arange(u.shape[0])

    x = data.scales.x.offset + (x+.5)*data.scales.x.slope*data.grid.x
    y = data.scales.y.offset + (y+.5)*data.scales.y.slope*data.grid.y

    x,y = np.meshgrid(x,y)
    dataset = from_arrays(x,y,u,v,mask,frame=frame)

    dataset["t"].assign_coords({"t":dataset.t+frame})

    dataset.attrs["files"].append(str(filename))
    dataset.attrs["delta_t"]  = data.attributes['FrameDt']

    return dataset

def load_directory(
        path: pathlib.Path, 
        basename: str="*", 
        ext: str=".vec",
        )->xr.Dataset:
    """
    load_directory (path,basename='*', ext='*.vec')

    Loads all the files with the chosen sextension in the directory into a
    single xarray Dataset with variables and units added as attributes

    Input:
        directory : path to the directory with .vec, .txt or .VC7 files,
        period . can be dropped

    Output:
        dataset : xarray Dataset with dimensions: x,y,t and
               dataset arrays of u,v,
               attributes of variables and units


    See more: load_vec
    """

    files = sorted(path.glob(basename+ext))


    if len(files) == 0:
        raise IOError(f"No files {basename+ext} in the directory {path} ")
    else:
        print(f"found {len(files)} files")

    dataset = []
    combined = []

    variables, units, rows, cols, dt, frame, method = parse_header(files[0])

    if method == load_vc7:
        for i, f in enumerate(files):
                dataset.append(load_vc7(f, frame=i))

    else:
        for i,f in enumerate(files):
            dataset.append(
                method(
                    f,
                    rows=rows,
                    cols=cols,
                    frame=i,
                    dt=dt
                )
            )

    if len(dataset) > 0:
        combined = xr.concat(dataset, dim="t")
        combined.attrs["delta_t"] = dataset[-1].attrs["delta_t"]
        combined.attrs["files"] = str(files)
        return combined
                        
    else:
        raise IOError("Could not read the files")

    


def parse_header(filename: pathlib.Path)-> Tuple[str, ...]:
    """
    parse_header ( filename)
    Parses header of the file (.vec) to get the variables (typically X,Y,U,V)
    and units (can be m,mm, pix/DELTA_T or mm/sec, etc.), and the size of the
    Dataset by the number of rows and columns.
    Input:
        filename : complete path of the file to read, pathlib.Path
    Returns:
        variables : list of strings
        units : list of strings
        rows : number of rows of the Dataset
        cols : number of columns of the Dataset
        dt   : time interval between the two PIV frames in microseconds
        method: function to load
    """

    # split path from the filename
    fname = str(filename.name).split('.')[0]

    try:
        frame = int(re.findall(r'\d+', fname)[-1])
        # print(int(re.findall(r'\d+', tmp)[-1]))
        # print(int(''.join(filter(str.isdigit,tmp))[-1]))
        # print(int(re.findall(r'[0-9]+', tmp)[-1]))        
    except:
        frame = 0

    # binary, no header
    if filename.suffix.lower() == '.vc7':
        return (
            ["x", "y", "u", "v"],
            4*[POS_UNITS],
            None,
            None,
            None,
            frame,
            load_vc7,
        )
    
    with open(filename,"r") as fid:
        header = fid.readline()
        # print(header)

    # if the file does not have a header, can be from OpenPIV or elsewhere
    # return None
    if header.startswith('#DaVis'):
        header_list = header.split(' ')
        rows = header_list[4]
        cols = header_list[5]
        pos_units = header_list[7]
        vel_units = header_list[-1]
        variables = ['x','y','u','v']
        units = [pos_units, pos_units, vel_units, vel_units]
        dt = 0.0
        method = load_davis8_txt
        return variables, units, rows, cols, dt, frame, method

    elif header.startswith('TITLE='): # Insight
        header_list = (
            header.replace(",", " ").replace("=", " ").replace('"', " ").split()
        )

        # get variable names, typically X,Y,U,V
        variables = header_list[3:12][::2]

        # get units - this is important if it's mm or m/s
        units = header_list[4:12][::2]

        # get the size of the PIV grid in rows x cols
        rows = int(header_list[-5])
        cols = int(header_list[-3])

        # this is also important to know the time interval, DELTA_T
        ind1 = header.find("MicrosecondsPerDeltaT")
        dt = float(header[ind1:].split('"')[1])            
        method = load_vec

        return variables, units, rows, cols, dt, frame, method

    else: # no header, probably OpenPIV txt
        method = load_openpiv_txt
        return (
            ["x", "y", "u", "v"],
            4*[POS_UNITS],
            None,
            None,
            None,
            frame,
            method,
        )


def get_units(filename: pathlib.Path)->Tuple[str,str,float]:
    """
    get_units(filename)

    given a full path name to the .vec file will return the names
    of length and velocity units fallback option is all None. Uses
    parse_header function, see below.

    """

    _, units, _, _, _, _,_ = parse_header(filename)

    if units == "":
        return (POS_UNITS, VEL_UNITS, DELTA_T)

    lUnits = units[0] # either m, mm, pix
    velUnits = units[2] # either m/s, mm/s, pix

    tUnits = velUnits.split("/")[1]  # make it 's' if exists

    return (lUnits, velUnits, tUnits)


def load_openpiv_txt(
    filename: str,
    rows: int=None,
    cols: int=None,
    dt: float=None,
    frame: int=0,
)-> xr.Dataset:
    """
        load_vec(filename,rows=rows,cols=cols)
        Loads the VEC file (TECPLOT format by TSI Inc.), OpenPIV VEC or TXT
        formats
        Arguments:
            filename : file name, expected to have a header and 5 columns
            rows, cols : number of rows and columns of a vector field,
            if None, None, then parse_header is called to infer the number
            written in the header
            DELTA_T : time interval (default is None)
            frame : frame or time marker (default is None)
        Output:
            dataset is a xAarray Dataset, see xarray for help
    """
    if rows is None:  # means no headers
        d = np.genfromtxt(filename, usecols=(0, 1, 2, 3, 4))
        x = sorted_unique(d[:, 0])
        y = sorted_unique(d[:, 1])
        d = d.reshape((len(y), len(x), 5)).transpose(1, 0, 2)
    else:
        d = np.genfromtxt(
            filename, skip_header=1, delimiter=",", usecols=(0, 1, 2, 3, 4)
        )
        d = d.reshape((rows, cols, 5))

        x = d[:, :, 0][0, :]
        y = d[:, :, 1][:, 0]

    u = d[:, :, 2]
    v = d[:, :, 3]
    chc = d[:, :, 4]


    dataset = xr.Dataset({
        "u": xr.DataArray(
                    u[:, :, np.newaxis], 
                    dims=("x", "y", "t"), 
                    coords={"x": x, "y": y, "t": [frame]}
                    ), 
        "v": xr.DataArray(
                    v[:, :, np.newaxis], 
                    dims=("x", "y", "t"), 
                    coords={"x": x, "y": y, "t": [frame]}
                    ), 
        "chc": xr.DataArray(
                    chc[:, :, np.newaxis], 
                    dims=("x", "y", "t"), 
                    coords={"x": x, "y": y, "t": [frame]}
                    ),
        })

    dataset = set_default_attrs(dataset)
    if dt is not None:
        dataset.attrs["delta_t"] = dt
    dataset.attrs["files"].append(str(filename))

    return dataset


def load_davis8_txt(
    filename: pathlib.Path,
    rows: int=None,
    cols: int=None,
    dt: float=0.0,
    frame: int=0,
) -> xr.Dataset:
    df = pd.read_csv(
        filename,
        delimiter='\t',
        skiprows=1,
        names=['x','y','u','v'], 
        decimal=","
        )
    ds = from_df(df, frame=frame)
    return ds



def sorted_unique(array):
    """ Returns not sorted sorted_unique """
    uniq, index = np.unique(array, return_index=True)
    return uniq[index.argsort()]
